from functools import wraps
from time import sleep
from typing import List

from .rate_limit import RateLimit
from .exceptions import RateLimitHit


class OnHitAction:
    raise_exception = 0
    wait = 1


class RateLimiter:
    def __init__(self, storage=RateLimit, *, action=OnHitAction.raise_exception):
        self._limits = []
        self._storage = storage
        self.action = action

    def _create_single_limiter(self, name, callback, defaults=None):
        if defaults is None:
            defaults = [(), (), ()]
        self._limits.append(self._storage(name, callback, defaults))

    def create_limiter(self, names, callback, defaults=None):
        if isinstance(names, list):
            for name in names:
                self._create_single_limiter(name, callback, defaults)
        elif isinstance(names, str):
            self._create_single_limiter(names, callback, defaults)
        else:
            raise ValueError("names parameter must be either a string or a iterable")

    @staticmethod
    def can_request(limits):
        for limit in limits:
            print(limit, limit.can_request())
            if not limit.can_request():
                return [False, limit]
        return [True, None]

    @staticmethod
    def is_initialized(limits):
        for limit in limits:
            if not limit.is_initialized():
                return False
        return True

    @staticmethod
    def register_request(limits, rt):
        for limit in limits:
            limit.register_request(rt)

    def use(self, *limits_names):
        def request_wrapper(func):
            limits: List[RateLimit] = [limit for limit in self._limits if limit.name in limits_names]

            @wraps(func)
            def func_wrapper():
                rl = RateLimiter.can_request(limits)
                if rl[0]:
                    ret = func()
                    RateLimiter.register_request(limits, ret)
                    return ret
                else:
                    if self.action == OnHitAction.raise_exception:
                        raise RateLimitHit(rl[1])
                    elif self.action == OnHitAction.wait:
                        sleep(rl[1].time_until_new_request_is_possible)
                        ret = func()
                        RateLimiter.register_request(limits, ret)
                        return ret
            return func_wrapper
        return request_wrapper



from time import time
from typing import List


class LimitDto:
    def __init__(self, max_requests, requests_made, reset_timestamp):
        self.max_requests = max_requests
        self.requests_made = requests_made
        self.reset_timestamp = reset_timestamp

    def __repr__(self):
        return str(self.__dict__)


def define_limiter_headers(max_requests, requests_made, seconds_until_reset):
    def define_limiter_headers_wrapper():
        return LimitDto(
            max_requests,
            requests_made,
            reset_timestamp=int(time()) + seconds_until_reset
        )

    return define_limiter_headers_wrapper


class RateLimit:
    def __init__(self, name, formatter, defaults):
        self.name = name
        self.max_requests: List = list(defaults[0])
        self.requests_made: List = list(defaults[1])
        self.reset_timestamp: List = [time() + default for default in defaults[2]]

        self._formatter = formatter

    def __repr__(self):
        return str(self.__dict__)

    def can_request(self) -> bool:
        if not (self.max_requests and self.requests_made and self.reset_timestamp):
            return True
        for max_requests, requests_made in zip(self.max_requests, self.requests_made):
            if requests_made >= max_requests:
                return False
        return True

    def is_initialized(self):
        if self.max_requests and self.requests_made and self.reset_timestamp:
            return True
        return False

    def set_formatter(self, callback):
        self._formatter = callback

    def register_request(self, request=None, headers: dict = None):
        if None not in (request, headers):
            raise Exception("Only one argument can't be none")

        if request is not None:
            headers = request.headers

        limits: List[LimitDto] = self._formatter(headers)

        for idx, limit in enumerate(limits):
            if idx > len(self.requests_made)-1:
                self.requests_made.append(limit.requests_made)
                self.max_requests.append(limit.max_requests)
                self.reset_timestamp.append(limit.reset_timestamp)

            self.requests_made[idx] = limit.requests_made
            self.max_requests[idx] = limit.max_requests

            if limit.requests_made == 1:
                self.reset_timestamp[idx] = limit.reset_timestamp

    @property
    def reset_in(self):
        times = []
        for reset_time in self.reset_timestamp:
            times.append(reset_time - int(time()))
        return times

    @property
    def time_until_new_request_is_possible(self):
        for idx, (max_requests, requests_made) in enumerate(zip(self.max_requests, self.requests_made)):
            if requests_made >= max_requests:
                return self.reset_timestamp[idx] - int(time())



from setuptools import setup

__version__ = '0.0.1'

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='requests-limiter',
    version=__version__,
    author='André Sousa',
    author_email='andrematosdesousa@gmail.com',
    license='MIT',
    url='https://github.com/sousa-andre/requests-limiter/',
    long_description_content_type='text/markdown',
    long_description=long_description,
    packages=['limiter'],
    install_requires=[],
    project_urls={
        'Source': 'https://github.com/sousa-andre/requests-limiter/'
    }
)
